class ChannelId(str):

    # Prefix for meta channel IDs
    META = '/meta'

    # Connect meta channel ID
    META_CONNECT = META + '/connect'

    # Disconnect meta channel ID
    META_DISCONNECT = META + '/disconnect'

    # Handshake meta channel ID
    META_HANDSHAKE = META + '/handshake'

    # Publish meta channel ID
    META_PUBLISH = META + '/publish'

    # Subscribe meta channel ID
    META_SUBSCRIBE = META + '/subscribe'

    # Unsubscribe meta channel ID
    META_UNSUBSCRIBE = META + '/unsubscribe'

    # Unsuccessful meta channel ID
    META_UNSUCCESSFUL = META + '/unsuccessful'

    # Pattern for wildcard channel IDs
    WILD = '*'

    # Pattern for deep wildcard channel IDs
    WILD_DEEP = '**'

    def __new__(cls, o='', encoding='utf-8', errors='strict'):
        if isinstance(o, bytes):
            return super().__new__(cls, o, encoding=encoding, errors=errors)

        return super().__new__(cls, o)

    def __init__(self, *args, **kwargs):
        self._parts = self.split('/')

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        if not isinstance(other, str):
            raise TypeError('Expected %s, got %s' % self.__class__, type(other))
        return str.__eq__(self, other)

    @property
    def is_meta(self):
        return self.startswith(self.META)

    @property
    def is_wild(self):
        return self.endswith('/' + self.WILD)

    @property
    def is_wild_deep(self):
        return self.endswith('/' + self.WILD_DEEP)

    @property
    def parts(self):
        return self._parts[1:]

    def get_wilds(self):
        wilds = []
        parts = self._parts
        last_index = len(parts) - 1
        for index in range(last_index, 0, -1):
            name = '/'.join(parts[:index]) + '/*'
            if index == last_index:
                wilds.append(name)
            wilds.append(name + '*')
        return wilds

    @classmethod
    def convert(cls, value):
        if isinstance(value, cls):
            return value
        if not isinstance(value, str):
            raise TypeError('Expected string, got %s' % type(value))
        return cls(value)



import logging
from unittest import TestCase

from baiocas.extensions.base import Extension


class TestExtension(TestCase):

    def setUp(self):
        self.extension = Extension()

    def test_init(self):
        assert isinstance(self.extension.log, logging.Logger)
        assert self.extension.log.name == 'baiocas.extensions.base.Extension'
        assert self.extension.client is None

    def test_repr(self):
        assert repr(self.extension) == 'Extension'

    def test_name(self):
        assert self.extension.name == 'Extension'

    def test_receive(self):
        message = {'channel': '/test', 'id': '1'}
        new_message = self.extension.receive(message)
        assert new_message is message
        assert new_message == {'channel': '/test', 'id': '1'}

    def test_register(self):
        client = {}
        assert self.extension.client is None
        self.extension.register(client)
        assert self.extension.client is client

    def test_send(self):
        message = {'channel': '/test', 'id': '1'}
        new_message = self.extension.send(message)
        assert new_message is message
        assert new_message == {'channel': '/test', 'id': '1'}

    def test_unregister(self):
        client = {}
        assert self.extension.client is None
        self.extension.unregister()
        assert self.extension.client is None
        self.extension.register(client)
        assert self.extension.client is client
        self.extension.unregister()
        assert self.extension.client is None



import http.cookies
import time
from email.utils import mktime_tz
from email.utils import parsedate_tz

from baiocas.transports.base import Transport
from baiocas.transports.util import is_cookie_expired


class HttpTransport(Transport):

    DEFAULT_HEADERS = {
        'content-type': ['application/json; charset=UTF-8']
    }

    OPTION_HEADERS = 'request_headers'

    def __init__(self, *args, **kwargs):
        super(HttpTransport, self).__init__(*args, **kwargs)
        self._cookies = http.cookies.SimpleCookie()

    def add_header(self, name, value):
        headers = self._options.setdefault(self.OPTION_HEADERS, {})
        headers.setdefault(name.lower(), []).append(value)

    def configure(self, **options):
        if self.OPTION_HEADERS in options:
            headers = dict((name.lower(), values) for name, values
                           in options[self.OPTION_HEADERS].items())
            options[self.OPTION_HEADERS] = headers
        super(HttpTransport, self).configure(**options)

    def get_cookie(self, name):
        self.log.debug('Getting cookie with name "%s"' % name)
        return self._cookies.get(name)

    def get_cookie_headers(self, include_expired=False):
        cookies = []
        for key, cookie in sorted(self._cookies.items()):
            if include_expired or not is_cookie_expired(cookie):
                cookies.append(cookie.OutputString(attrs=[]))
        return cookies

    def get_headers(self):
        headers = self.DEFAULT_HEADERS.copy()
        headers.update(self._options.get(self.OPTION_HEADERS, {}))
        cookies = self.get_cookie_headers()
        if cookies:
            headers.setdefault('cookie', []).extend(cookies)
        return headers

    def remove_header(self, name, value):
        headers = self._options.get(self.OPTION_HEADERS, {})
        if name not in headers:
            return False
        del headers[name]
        return True

    def set_cookie(self, name, value, **attrs):
        self._cookies[name] = value
        cookie = self._cookies[name]
        cookie.update(attrs)
        cookie.time_received = time.time()
        self.log.debug('Set cookie %s = %s' % (name, value))
        return cookie

    def set_header(self, name, values):
        if not isinstance(values, (list, tuple)):
            values = [values]
        else:
            values = list(values)
        headers = self._options.setdefault(self.OPTION_HEADERS, {})
        headers[name.lower()] = values

    def update_cookies(self, values, time_received=None):
        cookies = http.cookies.SimpleCookie()
        if isinstance(time_received, (list, tuple)):
            time_received = time_received[0] if time_received else None
        if isinstance(time_received, str):
            time_received = parsedate_tz(time_received)
            if time_received:
                time_received = mktime_tz(time_received)
        if time_received is None:
            time_received = time.time()
        for value in values:
            cookies.load(value)
        for cookie in cookies.values():
            cookie.time_received = time_received
        self._cookies.update(cookies)
        self.log.debug('Updated cookie headers: %s' % values)



from baiocas.channel_id import ChannelId
from baiocas.extensions.base import Extension


class AckExtension(Extension):

    FIELD_ACK = 'ack'

    def __init__(self):
        super(AckExtension, self).__init__()
        self._server_supports_acks = False
        self._ack_id = None

    @property
    def ack_id(self):
        return self._ack_id

    @property
    def server_supports_acks(self):
        return self._server_supports_acks

    def _get_ack(self, message):
        return message.get(message.FIELD_EXT, {}).get(self.FIELD_ACK)

    def _set_ack(self, message, value):
        message.setdefault(message.FIELD_EXT, {})[self.FIELD_ACK] = value

    def receive(self, message):
        channel = message.channel
        if channel == ChannelId.META_HANDSHAKE:
            if self._get_ack(message):
                self._server_supports_acks = True
            self.log.debug('Server supports acks: %s' % self._server_supports_acks)
        elif self._server_supports_acks and channel == ChannelId.META_CONNECT and message.successful:
            if isinstance(self._get_ack(message), int):
                self._ack_id = self._get_ack(message)
                self.log.debug('Server sent ACK ID: %s' % self._ack_id)
        return message

    def send(self, message):
        channel = message.channel
        if channel == ChannelId.META_HANDSHAKE:
            self._set_ack(message, self._client.options.get('ack_enabled', True))
            self._ack_id = None
            self.log.debug('Handshake being sent, clearing ACK ID')
        elif self._server_supports_acks and channel == ChannelId.META_CONNECT:
            self._set_ack(message, self._ack_id)
            self.log.debug('Sending ACK ID: %s' % self._ack_id)
        return message



from email.utils import formatdate

from baiocas.extensions.base import Extension


class TimestampExtension(Extension):

    def send(self, message):
        message.timestamp = formatdate(usegmt=True)
        return message



from json import dumps
from json import loads

from baiocas.channel_id import ChannelId


class Message(dict):
    """
    For more information on subclassing dict so that all updates pass through
    __setitem__, see the following Stack Overflow post: http://bit.ly/dcoFkF.
    """

    FIELD_ADVICE = 'advice'

    FIELD_CHANNEL = 'channel'

    FIELD_CLIENT_ID = 'clientId'

    FIELD_CONNECTION_TYPE = 'connectionType'

    FIELD_DATA = 'data'

    FIELD_ERROR = 'error'

    FIELD_EXT = 'ext'

    FIELD_ID = 'id'

    FIELD_INTERVAL = 'interval'

    FIELD_MINIMUM_VERSION = 'minimumVersion'

    FIELD_RECONNECT = 'reconnect'

    FIELD_SUBSCRIPTION = 'subscription'

    FIELD_SUCCESSFUL = 'successful'

    FIELD_SUPPORTED_CONNECTION_TYPES = 'supportedConnectionTypes'

    FIELD_TIMEOUT = 'timeout'

    FIELD_TIMESTAMP = 'timestamp'

    FIELD_VERSION = 'version'

    RECONNECT_HANDSHAKE = 'handshake'

    RECONNECT_NONE = 'none'

    RECONNECT_RETRY = 'retry'

    def __init__(self, *args, **kwargs):
        for arg in [_f for _f in args if _f]:
            self.update(arg)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, name):
        key = self._get_key_from_name(name)
        if key:
            return self.get(key)
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        key = self._get_key_from_name(name)
        if key:
            self.__setitem__(key, value)
        else:
            object.__setattr__(self, name, value)

    def __setitem__(self, key, value):
        if key == self.FIELD_CHANNEL and value is not None:
            value = ChannelId.convert(value)
        dict.__setitem__(self, key, value)

    @property
    def failure(self):
        return not self.successful

    def _get_key_from_name(self, name):
        return getattr(self.__class__, 'FIELD_' + name.upper(), None)

    def copy(self):
        raw_message = super(Message, self).copy()
        return Message(raw_message)

    def setdefault(self, key, value=None):
        if key not in self:
            self[key] = value
        return self[key]

    def update(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError('update expected at most 1 arguments, got %d' % len(args))
        other = dict(*args, **kwargs)
        for key in other:
            self[key] = other[key]

    @classmethod
    def from_dict(cls, value):
        return Message(value)

    @classmethod
    def from_json(cls, value, encoding=None):
        if encoding is not None:
            value = value.decode(encoding, 'replace')
        messages = loads(value)
        if not isinstance(messages, (list, tuple)):
            messages = [messages]
        return list(map(cls.from_dict, messages))

    @classmethod
    def to_json(cls, messages, encoding=None):
        if not isinstance(messages, (list, tuple)):
            messages = [messages]
        value = dumps(messages, ensure_ascii=False)
        if encoding is not None:
            value = value.encode(encoding)
        return value


class FailureMessage(Message):

    FIELD_EXCEPTION = 'exception'

    FIELD_REQUEST = 'request'

    def __init__(self, *args, **kwargs):
        default_fields = dict(
            successful=False,
            exception=None,
            advice={
                self.FIELD_RECONNECT: self.RECONNECT_NONE,
                self.FIELD_INTERVAL: 0
            }
        )
        super(FailureMessage, self).__init__(default_fields, *args, **kwargs)

    @classmethod
    def from_message(cls, message, exception=None, **kwargs):
        return cls(
            id=message.id,
            channel=message.channel,
            request=message,
            exception=exception,
            **kwargs
        )



import time
from email.utils import formatdate
from http.cookies import Morsel
from unittest import TestCase

from baiocas.transports.util import is_cookie_expired


class TestExpiredCookie(TestCase):

    def setUp(self):
        self.cookie = Morsel()
        self.cookie.set('foo', 'bar', 'bar')
        self.cookie.time_received = time.time() - 1

    def test_no_expiration_information(self):
        assert not is_cookie_expired(self.cookie)

    def test_valid_max_age(self):
        self.cookie['max-age'] = '86400'
        assert not is_cookie_expired(self.cookie)

    def test_valid_max_age_missing_time_received(self):
        self.cookie['max-age'] = '1'
        del self.cookie.time_received
        assert not is_cookie_expired(self.cookie)

    def test_expired_max_age(self):
        self.cookie['max-age'] = '1'
        assert is_cookie_expired(self.cookie)
        self.cookie['max-age'] = '0'
        assert is_cookie_expired(self.cookie)

    def test_expired_max_age_missing_time_received(self):
        self.cookie['max-age'] = '0'
        del self.cookie.time_received
        assert is_cookie_expired(self.cookie)
        self.cookie['max-age'] = '1'
        assert not is_cookie_expired(self.cookie)

    def test_invalid_max_age(self):
        self.cookie['max-age'] = 'invalid'
        assert not is_cookie_expired(self.cookie)

    def test_valid_expires(self):
        self.cookie['expires'] = formatdate(time.time() + 86400, localtime=True)
        assert not is_cookie_expired(self.cookie)

    def test_expired_expires(self):
        self.cookie['expires'] = formatdate(usegmt=True)
        assert is_cookie_expired(self.cookie)

    def test_invalid_expires(self):
        self.cookie['expires'] = 'invalid'
        assert not is_cookie_expired(self.cookie)

    def test_valid_max_age_and_expired_expires(self):
        self.cookie['max-age'] = '86400'
        self.cookie['expires'] = formatdate(usegmt=True)
        assert not is_cookie_expired(self.cookie)

    def test_expired_max_age_and_valid_expires(self):
        self.cookie['max-age'] = '1'
        self.cookie['expires'] = formatdate(time.time() + 86400, localtime=True)
        assert is_cookie_expired(self.cookie)



import logging
import urllib.parse

from baiocas import errors
from baiocas.channel_id import ChannelId
from baiocas.message import Message


class Transport(object):

    DEFAULT_MAXIMUM_NETWORK_DELAY = 10000

    OPTION_MAXIMUM_NETWORK_DELAY = 'maximum_network_delay'

    def __init__(self, **options):
        self.log = logging.getLogger('%s.%s' % (self.__module__, self.name))
        self._client = None
        self._options = {}
        self.url = None
        self.configure(**options)

    def __repr__(self):
        return self.name

    @property
    def name(self):
        raise NotImplementedError('Must be implemented by child classes')

    @property
    def parsed_url(self):
        return self._parsed_url

    @property
    def options(self):
        return self._options.copy()

    def _get_url(self):
        return self._url

    def _set_url(self, url):
        parsed_url = urllib.parse.urlparse(url or '')
        if url is not None:
            if not parsed_url.hostname:
                raise errors.ConnectionStringError(url, self)
        self._url = url
        self._parsed_url = parsed_url

    url = property(_get_url, _set_url)

    def abort(self):
        self.log.debug('Transport aborted')

    def accept(self, bayeux_version):
        raise NotImplementedError('Must be implemented by child classes')

    def configure(self, **options):
        if not options:
            return
        self._options.update(options)
        self.log.debug('Options changed to: %s' % self._options)

    def get_timeout(self, messages):
        timeout = self._options.get(
            self.OPTION_MAXIMUM_NETWORK_DELAY,
            self.DEFAULT_MAXIMUM_NETWORK_DELAY
        )
        if len(messages) == 1 and messages[0].channel == ChannelId.META_CONNECT:
            advice = messages[0].advice
            if not advice or Message.FIELD_TIMEOUT not in advice:
                advice = self._client.advice
            timeout += advice[Message.FIELD_TIMEOUT]
        return timeout

    def register(self, client, url=None):
        self.log.debug('Executing registration callback')
        self._client = client
        self.url = url

    def reset(self):
        self.log.debug('Transport reset')

    def send(self, messages, sync=False):
        raise NotImplementedError('Must be implemented by child classes')

    def unregister(self):
        self.log.debug('Executing unregistration callback')
        self._client = None
        self.url = None



from email.utils import formatdate
from unittest import TestCase

from mock import patch

from baiocas.client import Client
from baiocas.extensions import timestamp
from baiocas.message import Message


class TestTimestampExtension(TestCase):

    def setUp(self):
        self.extension = timestamp.TimestampExtension()
        self.client = Client('http://www.example.com')
        self.extension.register(self.client)

    def test_receive(self):
        message = Message(channel='/test')
        assert self.extension.receive(message) is message
        assert message == {'channel': '/test'}

    def test_send(self):
        message = Message(channel='/test')
        with patch.object(timestamp, 'formatdate') as mock_formatdate:
            mock_formatdate.return_value = formatdate(usegmt=True)
            assert self.extension.send(message) is message
        mock_formatdate.assert_called_with(usegmt=True)
        assert message.timestamp == mock_formatdate.return_value



from tornado import gen
from tornado.httpclient import AsyncHTTPClient
from tornado.httpclient import HTTPClient
from tornado.httpclient import HTTPError
from tornado.httpclient import HTTPRequest
from tornado.httputil import HTTPHeaders

from baiocas import errors
from baiocas.message import Message
from baiocas.transports.http import HttpTransport


class LongPollingHttpTransport(HttpTransport):

    def __init__(self, *args, **kwargs):
        super(LongPollingHttpTransport, self).__init__(*args, **kwargs)
        self._append_message_type = False
        self._create_http_clients()

    @property
    def name(self):
        return 'long-polling'

    def _reset_http_clients(self):
        self._http_client.close()
        self._blocking_http_client.close()
        self._create_http_clients()

    def _create_http_clients(self):
        try:
            AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")
        # Can't use ModuleNotFoundError because it was added in Python 3.6
        except ImportError:
            pass
        self._http_client = AsyncHTTPClient()
        self._blocking_http_client = HTTPClient(
            async_client_class=AsyncHTTPClient
        )

    def _handle_response(self, response, messages):

        # Log the received response code and headers
        self.log.debug('Received response: %s' % response.code)
        for header, value in response.headers.get_all():
            self.log.debug('Response header %s: %s' % (header, value))

        # If there was an error, report the sent messages as failed
        if response.error:
            if isinstance(response.error, HTTPError):
                if response.error.code == 599:
                    error = errors.TimeoutError()
                else:
                    error = errors.ServerError(response.error.code)
            else:
                error = errors.CommunicationError(response.error)
            self.log.debug('Failed to send messages: %s' % error)
            self._client.fail_messages(messages, error)
            return

        # Update the cookies
        self.update_cookies(
            response.headers.get_list('Set-Cookie'),
            time_received=response.headers.get('Date')
        )

        # Get the received messages
        self.log.debug('Received body: %s' % response.body)
        messages = Message.from_json(response.body, encoding='utf8')
        self._client.receive_messages(messages)

    def _prepare_request(self, messages):

        # Determine the URL for the messages
        url = self.url
        if self._append_message_type and len(messages) == 1 and messages[0].channel.is_meta():
            message_type = '/'.join(messages[0].channel.parts()[1:])
            if not url.endswith('/'):
                url += '/'
            url += message_type

        # Get the headers for the request
        headers = HTTPHeaders()
        for header, values in self.get_headers().items():
            for value in values:
                headers.add(header, value)
        for header, value in headers.get_all():
            self.log.debug('Request header %s: %s' % (header, value))

        # Get the body for the request
        body = Message.to_json(messages, encoding='utf8')
        self.log.debug('Request body (length: %d): %s' % (len(body), body))

        # Get the timeout (in seconds)
        timeout = self.get_timeout(messages) / 1000.0
        self.log.debug('Request timeout: %ss' % timeout)

        # Build and return the request
        return HTTPRequest(
            url,
            method='POST',
            headers=headers,
            body=body,
            connect_timeout=timeout,
            request_timeout=timeout
        )

    def _set_url(self, url):
        super(LongPollingHttpTransport, self)._set_url(url)
        self._append_message_type = (
            len(self.parsed_url.scheme.strip()) > 0 and
            len(self.parsed_url.netloc.strip()) > 0 and
            len(self.parsed_url.query.strip()) == 0 and
            len(self.parsed_url.fragment.strip()) == 0
        )

    def abort(self):
        super(LongPollingHttpTransport, self).abort()
        self.log.debug('Cancelling pending requests')
        self._reset_http_clients()

    def accept(self, bayeux_version):
        return True

    @gen.coroutine
    def send(self, messages, sync=False):
        request = self._prepare_request(messages)

        # Send the message. For synchronous requests, the Tornado client doesn't
        # return the actual response object if there is an error even though it
        # generates it internally. Instead of recreating it, we'll grab it.
        self.log.debug('Sending message to %s' % request.url)
        if sync:
            try:
                response = self._blocking_http_client.fetch(request)
            except HTTPError:
                response = self._blocking_http_client._response
        else:
            response = yield self._http_client.fetch(request)

        # Handle the response. We catch all exceptions here so that a bad
        # response doesn't end up crashing the Tornado async framework.
        try:
            self._handle_response(response, messages)
        except Exception as ex:
            error = errors.CommunicationError(ex)
            self.log.debug('Exception handling response: %s' % error)
            self._client.fail_messages(messages, error)



import logging

from baiocas.channel_id import ChannelId
from baiocas.listener import Listener
from baiocas.message import Message


class Channel(object):

    def __init__(self, client, channel_id):
        self.log = logging.getLogger('%s.%s' % (self.__module__, self.__class__.__name__))
        self._client = client
        self._channel_id = ChannelId.convert(channel_id)
        self._listener_id = 0
        self._listeners = []
        self._subscriptions = []

    def __repr__(self):
        return self._channel_id

    @property
    def channel_id(self):
        return self._channel_id

    @property
    def is_meta(self):
        return self._channel_id.is_meta

    @property
    def is_wild(self):
        return self._channel_id.is_wild

    @property
    def is_wild_deep(self):
        return self._channel_id.is_wild_deep

    @property
    def parts(self):
        return self._channel_id.parts

    @property
    def has_subscriptions(self):
        return len(self._subscriptions) > 0

    def _add_listener(self, listeners, function, extra_args, extra_kwargs):
        self._listener_id += 1
        listeners.append(Listener(
            id=self._listener_id,
            function=function,
            extra_args=extra_args,
            extra_kwargs=extra_kwargs
        ))
        self.log.debug('Added listener "%s" for channel %s' %
                       (function.__name__, self._channel_id))
        return self._listener_id

    def _notify_listeners(self, listeners, channel, message):
        for listener in listeners:
            try:
                self.log.debug('Notifying listener "%s" of message' % listener.function.__name__)
                listener.function(channel, message, *listener.extra_args, **listener.extra_kwargs)
            except Exception as ex:
                self.log.warning('Exception with listener "%s" with %s: %s' %
                                 (listener.function.__name__, message, ex))
                self._client.fire(self._client.EVENT_LISTENER_EXCEPTION, listener, message, ex)

    def _remove_listener(self, listeners, id=None, function=None):
        if (id is not None) == (function is not None):
            raise ValueError('Either id or function must be given')
        success = False
        if id is not None:
            for index, listener in enumerate(listeners):
                if listener.id == id:
                    function = listener.function
                    del listeners[index]
                    success = True
                    break
        else:
            to_remove = []
            for index, listener in enumerate(listeners):
                if function == listener.function:
                    to_remove.append(index)
            for index in reversed(to_remove):
                del listeners[index]
            success = bool(to_remove)
        if success:
            self.log.debug('Removed listener(s) "%s" for channel %s' %
                           (function.__name__, self._channel_id))
        return success

    def add_listener(self, function, *extra_args, **extra_kwargs):
        return self._add_listener(self._listeners, function, extra_args, extra_kwargs)

    def clear_listeners(self):
        self._listeners = []
        self.log.debug('Cleared listeners for channel %s' % self._channel_id)

    def clear_subscriptions(self):
        self._subscriptions = []
        self.log.debug('Cleared subscriptions for channel %s' % self._channel_id)

    def get_wilds(self):
        return self._channel_id.get_wilds()

    def notify_listeners(self, channel, message):
        self._notify_listeners(self._listeners, channel, message)
        if message.data:
            self._notify_listeners(self._subscriptions, channel, message)

    def publish(self, data, properties=None):
        self.log.debug('Publishing data to channel: %s' % data)
        message = Message(properties, channel=self._channel_id, data=data)
        self._client.send(message)

    def remove_listener(self, id=None, function=None):
        return self._remove_listener(self._listeners, id=id, function=function)

    def subscribe(self, function, *extra_args, **extra_kwargs):
        properties = None
        if 'properties' in extra_kwargs:
            properties = extra_kwargs.pop('properties')
        if not self.has_subscriptions:
            self.log.debug('Subscribe to channel "%s"' % self._channel_id)
            message = Message(properties,
                              channel=ChannelId.META_SUBSCRIBE,
                              subscription=self._channel_id
                              )
            self._client.send(message)
        return self._add_listener(self._subscriptions, function, extra_args, extra_kwargs)

    def unsubscribe(self, id=None, function=None, properties=None):
        success = self._remove_listener(self._subscriptions, id=id, function=function)
        if not self.has_subscriptions:
            self.log.debug('Channel has no remaining subscriptions, sending unsubscribe')
            message = Message(properties,
                              channel=ChannelId.META_UNSUBSCRIBE,
                              subscription=self._channel_id
                              )
            self._client.send(message)
        return success



import logging
from contextlib import contextmanager
from datetime import timedelta

from tornado.ioloop import IOLoop

from baiocas import errors
from baiocas.channel import Channel
from baiocas.channel_id import ChannelId
from baiocas.listener import Listener
from baiocas.message import FailureMessage
from baiocas.message import Message
from baiocas.status import ClientStatus
from baiocas.transports.long_polling import LongPollingHttpTransport
from baiocas.transports.registry import TransportRegistry


class Client(object):

    BAYEUX_VERSION = '1.0'

    DEFAULT_OPTIONS = {
        'backoff_period_increment': 1000,
        'maximum_backoff_period': 60000,
        'reverse_incoming_extensions': True,
        'advice': {
            Message.FIELD_TIMEOUT: 60000,
            Message.FIELD_INTERVAL: 0,
            Message.FIELD_RECONNECT: Message.RECONNECT_RETRY
        }
    }

    EVENT_EXTENSION_EXCEPTION = 'extension_exception'

    EVENT_LISTENER_EXCEPTION = 'listener_exception'

    MINIMUM_BAYEUX_VERSION = '0.9'

    def __init__(self, url, **options):

        # Set up a logger for the client
        self.log = logging.getLogger('%s.%s' % (self.__module__, self.__class__.__name__))

        # Save the URL
        self._url = url

        # Use the default IO loop
        self.io_loop = IOLoop.current()

        # Current client status and connection properties
        self._status = ClientStatus.UNCONNECTED
        self._connected = False
        self._scheduled_send = None
        self._backoff_period = 0
        self._advice = {}

        # Channels keyed by channel ID
        self._channels = {}

        # Active connection properties
        self._client_id = None
        self._message_id = 0

        # Transport registry and current transport
        self._transports = TransportRegistry()
        self._transport = None

        # Keep track of batches, both manual and internal
        self._batch_id = 0
        self._internal_batch = False
        self._message_queue = []

        # Extensions
        self._extensions = []

        # Event listeners keyed by event
        self._event_listener_id = 0
        self._event_listeners = {}

        # Configure the client
        self._options = self.DEFAULT_OPTIONS.copy()
        self.configure(**options)

    @property
    def advice(self):
        return self._advice.copy()

    @property
    def backoff_period(self):
        return self._backoff_period

    @property
    def client_id(self):
        return self._client_id

    @property
    def is_batching(self):
        return self._batch_id > 0 or self._internal_batch

    @property
    def is_disconnected(self):
        return ClientStatus.is_disconnected(self._status)

    @property
    def message_id(self):
        return self._message_id

    @property
    def options(self):
        return self._options.copy()

    @property
    def status(self):
        return self._status

    @property
    def transport(self):
        return self._transport

    @property
    def url(self):
        return self._url

    def _apply_extension(self, extension, message, outgoing=False):
        try:
            self.log.debug('Applying extension %s to message' % extension)
            method = outgoing and 'send' or 'receive'
            return getattr(extension, method)(message)
        except Exception as ex:
            self.log.warning('Exception during execution of extension %s: %s' %
                             (extension, ex))
            self.fire(self.EVENT_EXTENSION_EXCEPTION, message, ex, outgoing=outgoing)

    def _apply_incoming_extensions(self, message):
        self.log.debug('Applying extensions to incoming message')
        extensions = self._extensions
        if self._options['reverse_incoming_extensions']:
            self.log.debug('Reversing order of extensions')
            extensions = reversed(self._extensions)
        for extension in extensions:
            message = self._apply_extension(extension, message)
            if not message:
                self.log.debug('Message cancelled, skipping other extensions')
                break
        return message

    def _apply_outgoing_extensions(self, message):
        self.log.debug('Applying extensions to outgoing message')
        for extension in self._extensions:
            message = self._apply_extension(extension, message, outgoing=True)
            if not message:
                self.log.debug('Message cancelled, skipping other extensions')
                break
        return message

    def _cancel_delayed_send(self):
        if not self._scheduled_send:
            return
        if self._scheduled_send:
            self.log.debug('Cancelling delayed send')
            self.io_loop.remove_timeout(self._scheduled_send)
        self._scheduled_send = None

    def _connect(self):

        # Don't attempt to connect if we're disconnected. This doesn't make much
        # sense, but that's how the JavaScript client is implemented.
        if self.is_disconnected:
            self.log.debug('Client is disconnected, skipping connect')
            return

        # Create the message. In case of a reload or temporary loss of
        # connection, we want the next successful connect to return immediately
        # instead of being held by the server so that listeners can be notified
        # that the connection has been re-established.
        message = Message(
            channel=ChannelId.META_CONNECT,
            connection_type=self._transport.name
        )
        if not self._connected:
            message[Message.FIELD_ADVICE] = {
                Message.FIELD_TIMEOUT: 0
            }

        # Connect
        self._set_status(ClientStatus.CONNECTING)
        self.log.debug('Sending connect: %s' % message)
        self._send(message, for_setup=True)
        self._set_status(ClientStatus.CONNECTED)

    def _delay_connect(self):
        self.log.debug('Scheduling delayed connect')
        self._set_status(ClientStatus.CONNECTING)
        self._delay_send(self._connect)

    def _delay_handshake(self):
        self.log.debug('Scheduling delayed handshake')
        self._set_status(ClientStatus.HANDSHAKING)
        self._internal_batch = True
        self._delay_send(self._handshake, properties=self._handshake_properties)

    def _delay_send(self, method, *args, **kwargs):
        self._cancel_delayed_send()
        delay = self._advice['interval'] + self._backoff_period
        self.log.debug('Send scheduled in %sms, interval = %s, backoff = %s: %s' %
                       (delay, self._advice['interval'], self._backoff_period, method.__name__))
        if delay == 0:
            method(*args, **kwargs)
        else:
            self._scheduled_send = self.io_loop.add_timeout(
                timedelta(seconds=delay / 1000.0),
                lambda: method(*args, **kwargs)
            )

    def _disconnect(self, abort=False):
        if self._status == ClientStatus.DISCONNECTED:
            return
        self.log.debug('Disconnecting client')
        self._set_status(ClientStatus.DISCONNECTED)
        self._cancel_delayed_send()
        if abort:
            self.log.debug('Aborting transport')
            self._transport.abort()
        self._client_id = None
        self._batch_id = 0
        self._reset_backoff_period()
        if len(self._message_queue) > 0:
            self.log.debug('Failing queued messages')
            self._handle_failure(self._message_queue[:], errors.StatusError(self._status))
            self._message_queue = []

    def _get_next_message_id(self):
        self._message_id += 1
        return self._message_id

    def _handle_connect_failure(self, message, exception):
        self.log.debug('Handling failed connect')
        self._connected = False
        self._notify_connect_failure(FailureMessage.from_message(
            message,
            exception=exception,
            advice={
                FailureMessage.FIELD_RECONNECT: FailureMessage.RECONNECT_RETRY,
                FailureMessage.FIELD_INTERVAL: self._backoff_period,
            },
        ))

    def _handle_connect_response(self, message):
        self.log.debug('Handling connect response')
        if self.is_disconnected:
            self.log.debug('Client disconnected, discarding connect response')
            return
        self._connected = message.successful
        if self._connected:
            self.log.info('Client is now connected')
            self._notify_listeners(ChannelId.META_CONNECT, message)
            action = self._advice[Message.FIELD_RECONNECT]
            if action == Message.RECONNECT_RETRY:
                self._reset_backoff_period()
                self._delay_connect()
            elif action == Message.RECONNECT_NONE:
                self._disconnect()
            else:
                raise errors.ActionError(action)
        else:
            self.log.info('Client failed to connect')
            self._notify_connect_failure(message)

    def _handle_disconnect_failure(self, message, exception):
        self.log.debug('Handling failed disconnect')
        self._notify_disconnect_failure(FailureMessage.from_message(message, exception=exception))

    def _handle_disconnect_response(self, message):
        self.log.debug('Handling disconnect response')
        if message.successful:
            self.log.info('Client is now disconnected')
            self._disconnect()
            self._notify_listeners(ChannelId.META_DISCONNECT, message)
        else:
            self.log.info('Client failed to disconnect')
            self._notify_disconnect_failure(message)

    def _handle_failure(self, messages, exception):
        self.log.debug('Handling %d failed messages for exception: %s' % (len(messages), exception))
        for message in messages:
            handler = self._handle_message_failure
            if message.channel.is_meta:
                handler_name = '_handle_%s_failure' % '_'.join(message.channel.parts[1:])
                self.log.debug('Looking for handler named %s' % handler_name)
                if hasattr(self, handler_name):
                    handler = getattr(self, handler_name)
            self.log.debug('Passing message to handler %s' % handler.__name__)
            handler(message, exception)

    def _handle_handshake_failure(self, message, exception):
        self.log.debug('Handling failed handshake')
        self._notify_handshake_failure(FailureMessage.from_message(
            message,
            exception=exception,
            advice={
                FailureMessage.FIELD_RECONNECT: FailureMessage.RECONNECT_RETRY,
                FailureMessage.FIELD_INTERVAL: self._backoff_period,
            },
        ))

    def _handle_handshake_response(self, message):

        # Fail immediately if the message was not successful
        self.log.debug('Handling handshake response')
        if not message.successful:
            self.log.info('Client failed to handshake')
            self._notify_handshake_failure(message)
            return

        # Save the client ID
        self.log.info('Handshake successful, client ID = %s' % message.client_id)
        self._client_id = message.client_id

        # Negotiate a transport with the server
        new_transport = self._transports.negotiate_transport(
            message.supported_connection_types or [],
            message.version
        )
        if not new_transport:
            raise errors.TransportNegotiationError(
                self._transports.find_transports(message.version),
                message.supported_connection_types or []
            )
        elif self._transport != new_transport:
            self.log.debug('Transport %s -> %s' % (self._transport, new_transport))
            self._transport = new_transport

        # The new transport is now in place, so the listeners can perform a
        # publish() if they want. Notify the listeners of the connect below.
        self._notify_listeners(ChannelId.META_HANDSHAKE, message)

        # Handle the advice action
        action = self._advice[Message.FIELD_RECONNECT]
        if self.is_disconnected:
            action = Message.RECONNECT_NONE
        if action == Message.RECONNECT_RETRY:
            self._reset_backoff_period()
            self._delay_connect()
        elif action == Message.RECONNECT_NONE:
            self._disconnect()
        else:
            raise errors.ActionError(action)

        # End the internal batch and allow held messages from the application to
        # go to the server (see _handshake() where we start the internal batch).
        self._internal_batch = False
        self.flush_batch()

    def _handle_message_failure(self, message, exception):
        self.log.debug('Handling failed message')
        self._notify_message_failure(FailureMessage.from_message(message, exception=exception))

    def _handle_message_response(self, message):
        self.log.debug('Handling message response')
        if message.successful is None:
            self.log.debug('Client received message with blank successful flag')
            if message.data:
                self._notify_listeners(message.channel, message)
            else:
                self.log.warning('Unknown message received: %s' % message)
        elif message.successful is True:
            self.log.debug('Client received successful message')
            self._notify_listeners(ChannelId.META_PUBLISH, message)
        else:
            self.log.debug('Client received unsuccessful message')
            self._notify_message_failure(message)

    def _handle_subscribe_failure(self, message, exception):
        self.log.debug('Handling failed subscribe')
        self._notify_subscribe_failure(FailureMessage.from_message(message, exception=exception))

    def _handle_subscribe_response(self, message):
        self.log.debug('Handling subscribe response')
        channel = message.subscription
        if message.successful:
            self.log.info('Client subscribed to channel "%s"' % channel)
            self._notify_listeners(ChannelId.META_SUBSCRIBE, message)
        else:
            self.log.info('Client failed to subscribe to channel "%s"' % channel)
            self._notify_subscribe_failure(message)

    def _handle_unsubscribe_failure(self, message, exception):
        self.log.debug('Handling failed unsubscribe')
        self._notify_unsubscribe_failure(FailureMessage.from_message(message, exception=exception))

    def _handle_unsubscribe_response(self, message):
        self.log.debug('Handling unsubscribe response')
        channel = message.subscription
        if message.successful:
            self.log.info('Client unsubscribed from channel "%s"' % channel)
            self._notify_listeners(ChannelId.META_UNSUBSCRIBE, message)
        else:
            self.log.info('Client failed to unsubscribe from channel "%s"' % channel)
            self._notify_unsubscribe_failure(message)

    def _handshake(self, properties=None):

        # Reset state before starting
        self.log.info('Starting handshake')
        self._client_id = None
        self.clear_subscriptions()

        # Reset the transports if we're not retrying the handshake. If we are
        # retrying the handshake, either because another handshake failed and
        # we're backing off or because the server timed us out and asked us to
        # re-handshake, make sure that the next action is a connect if the
        # handshake succeeds.
        if self.is_disconnected:
            self.log.debug('Client disconnected, resetting advice')
            self._transports.reset()
            self._update_advice(self._options['advice'])
        else:
            self.log.debug('Client not disconnected, using retry advice')
            advice = self._advice.copy()
            advice[Message.FIELD_RECONNECT] = Message.RECONNECT_RETRY
            self._update_advice(advice)

        # Mark the start of an internal batch. Since all calls are asynchronous,
        # this ensures that no other messages are sent until the connection is
        # fully established.
        self._batch_id = 0
        self._internal_batch = True

        # Save the properties provided so we can reuse them during re-handshakes
        self._handshake_properties = properties

        # Figure out the tranpsorts to send to the server
        transport_names = self._transports.find_transports(self.BAYEUX_VERSION)
        self.log.debug('Supported transports: %s' % ', '.join(transport_names))

        # Create the handshake message
        message = Message(properties,
                          version=self.BAYEUX_VERSION,
                          minimum_version=self.MINIMUM_BAYEUX_VERSION,
                          channel=ChannelId.META_HANDSHAKE,
                          supported_connection_types=transport_names,
                          advice={
                              Message.FIELD_TIMEOUT: self._advice[Message.FIELD_TIMEOUT],
                              Message.FIELD_INTERVAL: self._advice[Message.FIELD_INTERVAL]
                          }
                          )

        # Pick the first available transport as the initial transport since we
        # don't know what the server currently supports
        self._transport = self._transports.negotiate_transport(
            transport_names,
            self.BAYEUX_VERSION
        )
        self.log.debug('Initial transport: %s' % self._transport)

        # Bypass the internal batch and send immediately
        self._set_status(ClientStatus.HANDSHAKING)
        self.log.debug('Sending handshake: %s' % message)
        self._send(message, for_setup=True)

    def _increase_backoff_period(self):
        if self._backoff_period < self._options['maximum_backoff_period']:
            self._backoff_period += self._options['backoff_period_increment']
            self.log.debug('Increasing backoff period to %s' % self._backoff_period)

    def _notify_connect_failure(self, message):
        self.log.debug('Notifying listeners of failed connect')
        self._notify_listeners(ChannelId.META_CONNECT, message)
        self._notify_listeners(ChannelId.META_UNSUCCESSFUL, message)
        action = self._advice[Message.FIELD_RECONNECT]
        if self.is_disconnected:
            action = Message.RECONNECT_NONE
        if action == Message.RECONNECT_RETRY:
            self.log.debug('Retry reconnect advice received')
            self._increase_backoff_period()
            self._delay_connect()
        elif action == Message.RECONNECT_HANDSHAKE:
            self.log.debug('Handshake reconnect advice received')
            self._transports.reset()
            self._reset_backoff_period()
            self._delay_handshake()
        elif action == Message.RECONNECT_NONE:
            self.log.debug('No reconnect advice received')
            self._disconnect()
        else:
            raise errors.ActionError(action)

    def _notify_disconnect_failure(self, message):
        self.log.debug('Notifying listeners of failed disconnect')
        self._disconnect(abort=True)
        self._notify_listeners(ChannelId.META_DISCONNECT, message)
        self._notify_listeners(ChannelId.META_UNSUCCESSFUL, message)

    def _notify_handshake_failure(self, message):
        self.log.debug('Notifying listeners of failed handshake')
        self._notify_listeners(ChannelId.META_HANDSHAKE, message)
        self._notify_listeners(ChannelId.META_UNSUCCESSFUL, message)
        if not self.is_disconnected and self._advice[Message.FIELD_RECONNECT] != Message.RECONNECT_NONE:
            self._increase_backoff_period()
            self._delay_handshake()
        else:
            self._disconnect()

    def _notify_listeners(self, channel_id, message):

        # Notify direct listeners
        self.log.debug('Notifying listeners for %s' % channel_id)
        channel = self.get_channel(channel_id)
        channel.notify_listeners(channel, message)

        # Notify the wildcard listeners
        for wild in channel.channel_id.get_wilds():
            self.log.debug('Notifying listeners for %s' % wild)
            wild_channel = self.get_channel(wild)
            wild_channel.notify_listeners(channel, message)

    def _notify_message_failure(self, message):
        self.log.debug('Notifying listeners of failed message')
        self._notify_listeners(ChannelId.META_PUBLISH, message)
        self._notify_listeners(ChannelId.META_UNSUCCESSFUL, message)

    def _notify_subscribe_failure(self, message):
        self.log.debug('Notifying listeners of failed subscribe')
        self._notify_listeners(ChannelId.META_SUBSCRIBE, message)
        self._notify_listeners(ChannelId.META_UNSUCCESSFUL, message)

    def _notify_unsubscribe_failure(self, message):
        self.log.debug('Notifying listeners of failed unsubscribe')
        self._notify_listeners(ChannelId.META_UNSUBSCRIBE, message)
        self._notify_listeners(ChannelId.META_UNSUCCESSFUL, message)

    def _queue_send(self, message):
        self.log.debug('Queueing message for sending: %s' % message)
        if self.is_batching or ClientStatus.is_handshaking(self._status):
            self.log.debug('In batch, adding message to queue')
            self._message_queue.append(message)
        else:
            self.log.debug('Sending message immediately')
            self._send(message)

    def _receive(self, message):
        self.log.debug('Receiving message: %s' % message)
        message = self._apply_incoming_extensions(message)
        if not message:
            self.log.debug('Message cancelled by extensions')
            return
        self._update_advice(message.advice)
        handler = self._handle_message_response
        if message.channel and message.channel.is_meta:
            handler_name = '_handle_%s_response' % '_'.join(message.channel.parts[1:])
            self.log.debug('Looking for handler named %s' % handler_name)
            if hasattr(self, handler_name):
                handler = getattr(self, handler_name)
        self.log.debug('Passing message to handler %s' % handler.__name__)
        handler(message)

    def _reset_backoff_period(self):
        self.log.debug('Resetting backoff period to 0')
        self._backoff_period = 0

    def _send(self, messages, for_setup=False, sync=False):

        # Make sure we got a list of messages
        if not isinstance(messages, (list, tuple)):
            if messages is None:
                messages = []
            else:
                messages = [messages]
        self.log.debug('Sending messages: %s' % messages)

        # Make sure we can send messages
        if not for_setup and self._status not in (ClientStatus.CONNECTING, ClientStatus.CONNECTED):
            self.log.debug('Client is not connected, cannot send messages')
            self._handle_failure(messages, errors.StatusError(self._status))
            return False

        # Prepare the messages by checking that they all have a client ID and
        # passing them through the outgoing extensions. We check for client IDs
        # since messages could have been generated before the handshake
        # completed and the client ID was known.
        prepared_messages = []
        for message in messages:
            if self._client_id:
                message['clientId'] = self._client_id
            message = self._apply_outgoing_extensions(message)
            if not message:
                continue
            message.id = str(self._get_next_message_id())
            prepared_messages.append(message)
        if not prepared_messages:
            self.log.debug('All messages cancelled by extensions, skipping send')
            return False

        # Pass off the messages to the transport
        self.log.debug('Prepared messages: %s' % prepared_messages)
        self._transport.send(prepared_messages, sync=sync)
        return True

    def _set_status(self, status):
        if status == self._status:
            return
        self.log.info('Status: %s -> %s' % (self._status, status))
        self._status = status

    def _update_advice(self, new_advice):
        if new_advice:
            advice = self._options['advice'].copy()
            advice.update(new_advice)
            self._advice = advice
            self.log.debug('New advice: %s' % self._advice)

    def clear_subscriptions(self):
        self.log.info('Clearing subscriptions')
        for channel in self._channels.values():
            channel.clear_subscriptions()

    def configure(self, **options):
        if not options:
            return
        self._options.update(options)
        self.log.debug('Options changed to: %s' % self._options)

    def disconnect(self, properties=None, sync=True):
        if self.is_disconnected:
            self.log.debug('Client already disconnected, skipping disconnect')
            return
        message = Message(properties, channel=ChannelId.META_DISCONNECT)
        self.log.debug('Sending disconnect: %s' % message)
        self._set_status(ClientStatus.DISCONNECTING)
        self._send(message, for_setup=True, sync=sync)

    def end_batch(self):
        if self._batch_id == 0:
            raise errors.BatchError()
        self.log.debug('Ended batch with ID %s' % self._batch_id)
        self._batch_id -= 1
        if not self.is_batching and not self.is_disconnected:
            self.flush_batch()

    def fail_messages(self, messages, exception=None):
        self.log.debug('Failing messages: %s' % messages)
        self._handle_failure(messages, exception)

    def fire(self, event, *args, **kwargs):
        self.log.debug('Firing event %s' % event)
        for listener in self._event_listeners.get(event, []):
            try:
                self.log.debug('Invoking callback "%s" for event %s' %
                               (listener.function.__name__, event))
                final_args = args
                final_kwargs = kwargs
                if listener.extra_args:
                    final_args += listener.extra_args
                if listener.extra_kwargs:
                    final_kwargs = final_kwargs.copy()
                    final_kwargs.update(listener.extra_kwargs)
                listener.function(self, *final_args, **final_kwargs)
            except Exception as ex:
                self.log.warning('Exception with listener "%s" for event %s: %s' %
                                 (listener.function.__name__, event, ex))

    def flush_batch(self):
        self.log.debug('Flushing batch of %d messages' % len(self._message_queue))
        if not self._message_queue:
            self.log.debug('No messages in batch queue, skipping flush')
            return
        messages = self._message_queue[:]
        self._message_queue = []
        self._send(messages)

    def get_channel(self, channel_id):
        self.log.debug('Fetching channel %s' % channel_id)
        channel_id = ChannelId.convert(channel_id)
        channel = self._channels.get(channel_id)
        if not channel:
            self.log.debug('Channel does not exist, creating with ID %s' % channel_id)
            channel = Channel(self, channel_id)
            self._channels[channel_id] = channel
        return channel

    def get_known_transports(self):
        self.log.debug('Fetching known transports')
        return self._transports.get_known_transports()

    def get_transport(self, name):
        self.log.debug('Getting transport with name "%s"' % name)
        return self._transports.get_transport(name)

    def handshake(self, properties=None):
        self.log.debug('Initiating client handshake')
        self._set_status(ClientStatus.DISCONNECTED)
        self._handshake(properties=properties)

    def initialize(self, properties=None, **options):
        self.log.debug('Initializing client with options: %s' % options)
        self.configure(**options)
        self.handshake(properties=properties)

    def receive_messages(self, messages):
        self.log.info('Received %d messages' % len(messages))
        list(map(self._receive, messages))

    def register_extension(self, extension):
        self._extensions.append(extension)
        self.log.debug('Registered extension %s' % extension)
        extension.register(self)
        return True

    def register_listener(self, event, function, *extra_args, **extra_kwargs):
        self.log.debug('Registered "%s" for event %s' % (function.__name__, event))
        self._event_listener_id += 1
        self._event_listeners.setdefault(event, []).append(Listener(
            id=self._event_listener_id,
            function=function,
            extra_args=extra_args,
            extra_kwargs=extra_kwargs
        ))
        return self._event_listener_id

    def register_transport(self, transport):
        if not self._transports.add(transport):
            self.log.warning('Failed to register transport %s' % transport)
            return False
        self.log.debug('Registered transport %s' % transport)
        transport.register(self, url=self._url)
        return True

    def send(self, message):
        self.log.debug('Received message for sending: %s' % message)
        self._queue_send(message)

    def start_batch(self):
        self._batch_id += 1
        self.log.debug('Started batch with ID %s' % self._batch_id)

    def unregister_extension(self, extension):
        if extension not in self._extensions:
            self.log.warning('Failed to unregister extension %s, not registered' % extension)
            return False
        self._extensions.remove(extension)
        extension.unregister()
        self.log.debug('Unregistered extension %s' % extension)
        return True

    def unregister_listener(self, id=None, event=None, function=None):
        if (id is not None) == (event is not None or function is not None):
            raise ValueError('Either id or event/function must be given')
        unregistered = 0
        if id is not None:
            for event, listeners in self._event_listeners.items():
                for index, listener in enumerate(listeners):
                    if listener.id == id:
                        del listeners[index]
                        unregistered = 1
                        break
        elif event is not None:
            listeners = self._event_listeners.get(event, [])
            to_remove = []
            for index, listener in enumerate(listeners):
                if function is None or listener.function == function:
                    to_remove.append(index)
            for index in reversed(to_remove):
                del listeners[index]
            unregistered += len(to_remove)
        else:
            for event, listeners in self._event_listeners.items():
                to_remove = []
                for index, listener in enumerate(listeners):
                    if listener.function == function:
                        to_remove.append(index)
                for index in reversed(to_remove):
                    del listeners[index]
                unregistered += len(to_remove)
        self.log.debug('Unregistered %d listeners' % unregistered)
        return bool(unregistered)

    def unregister_transport(self, name):
        transport = self._transports.remove(name)
        if not transport:
            self.log.warning('Failed to unregister transport %s, not registered' % name)
            return None
        self.log.debug('Unregistered transport %s' % transport)
        transport.unregister()
        return transport

    @contextmanager
    def batch(self):
        self.log.debug('Entered batch context manager')
        self.start_batch()
        try:
            self.log.debug('Waiting for batch content')
            yield
        finally:
            self.end_batch()
            self.log.debug('Exited batch context manager')


def get_client(url, **options):
    client = Client(url, **options)
    client.register_transport(LongPollingHttpTransport())
    return client



"""
Exceptions used with Baiocas.

All exceptions inherit from :class:`.BayeuxError`. Exceptions not inheriting
from this class are not raised directly by Baiocas.
"""


class BayeuxError(Exception):
    """Generic base error class for Baiocas."""

    def __eq__(self, other):
        if not isinstance(other, self.__class__) or \
                str(other) != str(self) or \
                other.args != self.args:
            return False
        for name, value in self.__dict__.items():
            if other.__dict__.get(name) != value:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class ActionError(BayeuxError):
    """Raised when an invalid advice action is encountered."""

    def __init__(self, action):
        message = 'Unrecognized advice action "%s"' % action
        super(ActionError, self).__init__(message)
        self.action = action


class BatchError(BayeuxError):
    """Raised when batches are not started/stopped in the right order."""


class CommunicationError(BayeuxError):
    """Raised when a communication error occurs with a transport."""

    def __init__(self, transport_exception):
        message = 'Communication error: %s' % transport_exception
        super(CommunicationError, self).__init__(message)
        self.transport_exception = transport_exception


class ConnectionStringError(BayeuxError):
    """Raised when an invalid connection string is encountered."""

    def __init__(self, transport, value):
        message = 'Invalid connection string, "%s", for transport %s' % (value, transport)
        super(ConnectionStringError, self).__init__(message)
        self.transport = transport
        self.value = value


class ServerError(BayeuxError):
    """Raised when a server responds with a non-successful status."""

    def __init__(self, code):
        message = 'Server responded with error %s' % code
        super(ServerError, self).__init__(message)
        self.code = code


class StatusError(BayeuxError):
    """Raised when the Bayeux client has an invalid status for an operation."""

    def __init__(self, status):
        message = 'Client status of "%s" is not valid for this operation' % status
        super(StatusError, self).__init__(message)
        self.status = status


class TimeoutError(BayeuxError):
    """Raised when the Bayeux client times out during an operation."""


class TransportNegotiationError(BayeuxError):
    """Raised when the Bayeux client and server could not agree on a transport."""

    def __init__(self, client_types, server_types):
        message = 'Could not negotiate transport with server; client: %s; server: %s' % \
            (', '.join(client_types), ', '.join(server_types))
        super(TransportNegotiationError, self).__init__(message)
        self.client_types = client_types
        self.server_types = server_types



from unittest import TestCase

from baiocas.transports.base import Transport
from baiocas.transports.registry import TransportRegistry


class MockTransport(Transport):

    def __init__(self, name, only_versions=None):
        self.__name = name
        self.__only_versions = only_versions
        self.is_reset = False
        super(MockTransport, self).__init__()

    @property
    def name(self):
        return self.__name

    def accept(self, version):
        if self.__only_versions is None:
            return True
        return version in self.__only_versions

    def reset(self):
        super(MockTransport, self).reset()
        self.is_reset = True


class TestTransportRegistry(TestCase):

    def setUp(self):
        self.registry = TransportRegistry()

    def test_add(self):
        transport = MockTransport('mock-transport')
        assert self.registry.add(transport)
        assert not self.registry.add(transport)
        assert transport.name in self.registry.get_known_transports()

    def test_find_transports(self):
        transport1 = MockTransport('mock-transport-1')
        transport2 = MockTransport('mock-transport-2', only_versions=['1.0'])
        self.registry.add(transport1)
        self.registry.add(transport2)
        transports = self.registry.find_transports('0.9')
        assert len(transports) == 1
        assert transport1.name in transports
        transports = self.registry.find_transports('1.0')
        assert len(transports) == 2
        assert transport1.name in transports
        assert transport2.name in transports

    def test_get_known_transports(self):
        assert len(self.registry.get_known_transports()) == 0
        transport1 = MockTransport('mock-transport-1')
        transport2 = MockTransport('mock-transport-2', only_versions=[])
        self.registry.add(transport1)
        self.registry.add(transport2)
        transports = self.registry.get_known_transports()
        assert len(transports) == 2
        assert transport1.name in transports
        assert transport2.name in transports

    def test_get_transport(self):
        transport = MockTransport('mock-transport')
        self.registry.add(transport)
        assert self.registry.get_transport(transport.name) is transport
        assert self.registry.get_transport('bad-transport') is None
        assert self.registry.get_transport(transport.name.upper()) is None

    def test_negotiate_transport(self):
        transport1 = MockTransport('mock-transport-1')
        transport2 = MockTransport('mock-transport-2', only_versions=['1.0'])
        self.registry.add(transport1)
        self.registry.add(transport2)
        assert self.registry.negotiate_transport([], '0.9') is None
        assert self.registry.negotiate_transport(['bad-transport'], '0.9') is None
        assert self.registry.negotiate_transport([transport1.name, transport2.name], '0.9') is transport1
        assert self.registry.negotiate_transport([transport2.name, transport1.name], '0.9') is transport1
        assert self.registry.negotiate_transport([transport1.name, transport2.name], '1.0') is transport1
        assert self.registry.negotiate_transport([transport2.name, transport1.name], '1.0') is transport2
        assert self.registry.remove(transport1.name) is transport1
        assert self.registry.negotiate_transport([transport1.name, transport2.name], '1.0') is transport2
        assert self.registry.negotiate_transport([transport1.name, transport2.name], '1.1') is None

    def test_remove(self):
        transport = MockTransport('mock-transport')
        assert self.registry.remove('bad-transport') is None
        self.registry.add(transport)
        assert len(self.registry.get_known_transports()) == 1
        assert self.registry.remove(transport.name.upper()) is None
        assert self.registry.remove(transport.name) is transport
        assert self.registry.remove(transport.name) is None
        assert len(self.registry.get_known_transports()) == 0

    def test_reset(self):
        transport1 = MockTransport('mock-transport-1')
        transport2 = MockTransport('mock-transport-2')
        self.registry.reset()
        self.registry.add(transport1)
        self.registry.add(transport2)
        assert not transport1.is_reset
        assert not transport2.is_reset
        self.registry.reset()
        assert transport1.is_reset
        assert transport2.is_reset



class ClientStatus(object):

    # State assumed after the handshake when the connection is broken
    UNCONNECTED = 'unconnected'

    # State assumed when the handshake is being sent
    HANDSHAKING = 'handshaking'

    # State assumed when a first handshake failed and the handshake is retried,
    # or when the Bayeux server requests a re-handshake
    REHANDSHAKING = 'rehandshaking'

    # State assumed when the connect is being sent for the first time
    CONNECTING = 'connecting'

    # State assumed when the client is connected to the Bayeux server
    CONNECTED = 'connected'

    # State assumed when the disconnect is being sent
    DISCONNECTING = 'disconnecting'

    # State assumed before the handshake and when the disconnect is completed
    DISCONNECTED = 'disconnected'

    @classmethod
    def is_disconnected(cls, status):
        return status in (cls.DISCONNECTING, cls.DISCONNECTED)

    @classmethod
    def is_handshaking(cls, status):
        return status in (cls.HANDSHAKING, cls.REHANDSHAKING)



from os import path

from setuptools import find_packages
from setuptools import setup

version = '2.0.0'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='baiocas',
    version=version,
    description='A simple Bayeux client for Python following the Comet paradigm.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    keywords='',
    author='Yoann Roman',
    author_email='silentsound@gmail.com',
    url='http://github.com/silentsound/baiocas',
    license_files=["LICENSE"],
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'tornado>=6.0'
    ],
    extras_require={
        'pycurl': ['pycurl>=7.22.0'],
    },
    entry_points='',
    python_requires='>=3.8',
)



import logging


class Extension(object):

    def __init__(self):
        self.log = logging.getLogger('%s.%s' % (self.__module__, self.name))
        self._client = None

    def __repr__(self):
        return self.name

    @property
    def client(self):
        return self._client

    @property
    def name(self):
        return self.__class__.__name__

    def receive(self, message):
        return message

    def register(self, client):
        self.log.debug('Executing registration callback')
        self._client = client

    def send(self, message):
        return message

    def unregister(self):
        self.log.debug('Executing unregistration callback')
        self._client = None



from unittest import TestCase

from baiocas.channel_id import ChannelId
from baiocas.client import Client
from baiocas.extensions.ack import AckExtension
from baiocas.message import Message


class TestAckExtension(TestCase):

    def setUp(self):
        self.extension = AckExtension()
        self.client = Client('http://www.example.com')
        self.extension.register(self.client)

    def test_init(self):
        assert self.extension.ack_id is None
        assert not self.extension.server_supports_acks

    def test_receive_handshake(self):
        message = Message(channel=ChannelId.META_HANDSHAKE)
        assert self.extension.receive(message) == message
        assert not self.extension.server_supports_acks
        message.ext = {AckExtension.FIELD_ACK: True}
        assert self.extension.receive(message) == message
        assert self.extension.server_supports_acks

    def test_receive_connect(self):

        # Check that nothing happens when no ACK ID is included
        message = Message(channel=ChannelId.META_CONNECT, successful=True)
        assert self.extension.receive(message) is message
        assert self.extension.ack_id is None
        assert not self.extension.server_supports_acks

        # Check that nothing happens when server support is unknown
        message.ext = {AckExtension.FIELD_ACK: 1}
        assert self.extension.receive(message) == message
        assert self.extension.ack_id is None
        assert not self.extension.server_supports_acks

        # Notify the extension that server supports ACKs
        self.extension.receive(Message(
            channel=ChannelId.META_HANDSHAKE,
            ext={AckExtension.FIELD_ACK: True}
        ))

        # Check that the ACK ID is captured
        assert self.extension.server_supports_acks
        assert self.extension.receive(message) == message
        assert self.extension.ack_id == 1

        # Check that the ACK ID is ignored for failed messages
        message.ext[AckExtension.FIELD_ACK] = 2
        message.successful = False
        assert self.extension.receive(message) == message
        assert self.extension.ack_id == 1

        # Check that the ACK ID is ignored if not an integer
        message.ext[AckExtension.FIELD_ACK] = '2'
        message.successful = True
        assert self.extension.receive(message) == message
        assert self.extension.ack_id == 1

        # Check that updates to the ACK ID are captured
        message.ext[AckExtension.FIELD_ACK] = 2
        assert self.extension.receive(message) == message
        assert self.extension.ack_id == 2

    def test_receive_other(self):
        message = Message(channel='/test', ext={AckExtension.FIELD_ACK: 1})
        assert self.extension.receive(message) is message
        assert not self.extension.server_supports_acks
        assert self.extension.ack_id is None

    def test_send_handshake(self):
        message = Message(channel=ChannelId.META_HANDSHAKE)
        assert self.extension.send(message) == message
        assert message.ext[AckExtension.FIELD_ACK]
        assert self.extension.ack_id is None
        self.client.configure(ack_enabled=False)
        assert self.extension.send(message) == message
        assert not message.ext[AckExtension.FIELD_ACK]
        assert self.extension.ack_id is None

    def test_send_connect(self):
        message = Message(channel=ChannelId.META_CONNECT)
        assert self.extension.send(message) == message
        assert not message.ext
        self.extension.receive(Message(
            channel=ChannelId.META_HANDSHAKE,
            ext={AckExtension.FIELD_ACK: True}
        ))
        assert self.extension.send(message) == message
        assert message.ext[AckExtension.FIELD_ACK] is None
        self.extension.receive(Message(
            channel=ChannelId.META_CONNECT,
            successful=True,
            ext={AckExtension.FIELD_ACK: 1}
        ))
        assert self.extension.send(message) == message
        assert message.ext[AckExtension.FIELD_ACK] == 1

    def test_send_other(self):
        message = Message(channel='/test')
        assert self.extension.send(message) == message
        assert message == {'channel': '/test'}



import time
from email.utils import mktime_tz
from email.utils import parsedate_tz


def is_cookie_expired(cookie):
    value = None
    if cookie['max-age']:
        value = cookie['max-age']
    elif cookie['expires']:
        value = cookie['expires']
    if not value:
        return False
    if value.isdigit():
        time_received = getattr(cookie, 'time_received', time.time())
        expires = time_received + int(value)
    else:
        expires = parsedate_tz(value)
        if expires:
            expires = mktime_tz(expires)
    if expires and expires <= time.time():
        return True
    return False



class TransportRegistry(object):

    def __init__(self):
        self._transports = {}

    def add(self, transport):
        if transport.name in self._transports:
            return False
        self._transports[transport.name] = transport
        return True

    def find_transports(self, version):
        return [transport.name for transport in list(self._transports.values())
                if transport.accept(version)]

    def get_known_transports(self):
        return list(self._transports.keys())

    def get_transport(self, name):
        return self._transports.get(name)

    def negotiate_transport(self, requested_transports, bayeux_version):
        for requested_transport in requested_transports:
            if requested_transport not in self._transports:
                continue
            transport = self._transports[requested_transport]
            if transport.accept(bayeux_version):
                return transport
        return None

    def remove(self, name):
        if name not in self._transports:
            return None
        transport = self._transports.pop(name)
        return transport

    def reset(self):
        for transport in list(self._transports.values()):
            transport.reset()
